# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json

import numpy as np
import pandas as pd
import pandas.core.common as pdc
import six
from jsontableschema import Schema
from jsontableschema.exceptions import InvalidObjectType


# Module API

def descriptor_and_rows_to_dataframe(descriptor, rows):
    # Prepare
    primary_key = None
    schema = Schema(descriptor)
    if len(schema.primary_key) == 1:
        primary_key = schema.primary_key[0]
    elif len(schema.primary_key) > 1:
        raise RuntimeError('Multi-column primary keys are not supported')

    # Get data/index
    data_rows = []
    index_rows = []
    jtstypes_map = {}
    for row in rows:
        values = []
        index = None
        for field, value in zip(schema.fields, row):
            try:
                value = field.cast_value(value)
            except InvalidObjectType:
                value = json.loads(value)
            if value is None and field.type in ('number', 'integer'):
                jtstypes_map[field.name] = 'number'
                value = np.NaN
            if field.name == primary_key:
                index = value
            else:
                values.append(value)
        data_rows.append(tuple(values))
        index_rows.append(index)

    # Get dtypes
    dtypes = []
    for field in schema.fields:
        if field.name != primary_key:
            field_name = field.name
            if six.PY2:
                field_name = field.name.encode('utf-8')
            dtype = jtstype_to_dtype(jtstypes_map.get(field.name, field.type))
            dtypes.append((field_name, dtype))

    # Create dataframe
    index = None
    columns = schema.headers
    array = np.array(data_rows, dtype=dtypes)
    if primary_key:
        index_field = schema.get_field(primary_key)
        index_dtype = jtstype_to_dtype(index_field.type)
        index_class = pd.Index
        if index_field.type in ['datetime', 'date']:
            index_class = pd.DatetimeIndex
        index = index_class(index_rows, name=primary_key, dtype=index_dtype)
        columns = filter(lambda column: column != primary_key, schema.headers)
    dataframe = pd.DataFrame(array, index=index, columns=columns)

    return dataframe


def dataframe_to_descriptor(dataframe):
    # Prepare
    fields = []
    primary_key = None

    # Primary key
    if dataframe.index.name:
        field_type = dtype_to_jtstype(dataframe.index.dtype)
        field = {
            'name': dataframe.index.name,
            'type': field_type,
            'constraints': {'required': True},
        }
        fields.append(field)
        primary_key = dataframe.index.name

    # Fields
    for column, dtype in dataframe.dtypes.iteritems():
        field_type = dtype_to_jtstype(dtype)
        field = {'name': column, 'type': field_type}
        if dataframe[column].isnull().sum() == 0:
            field['constraints'] = {'required': True}
        fields.append(field)

    # Descriptor
    descriptor = {}
    descriptor['fields'] = fields
    if primary_key:
        descriptor['primaryKey'] = primary_key

    return descriptor


def jtstype_to_dtype(jtstype):
    # Mapping
    MAPPING = {
        'string': np.dtype('O'),
        'number': np.dtype(float),
        'integer': np.dtype(int),
        'boolean': np.dtype(bool),
        'array': np.dtype(list),
        'object': np.dtype(dict),
        'date': np.dtype('datetime64[ns]'),
        'time': np.dtype('O'),
        'datetime': np.dtype('datetime64[ns]'),
        'year': np.dtype(int),
        'yearmonth': np.dtype(int),
        'geopoint': np.dtype('O'),
        'geojson': np.dtype('O'),
        'duration': np.dtype('O'),
        'any': np.dtype('O'),
    }

    # Get type
    try:
        dtype = MAPPING[jtstype]
    except KeyError:
        raise TypeError('JTS type "%s" is not supported' % jtstype)

    return dtype


def dtype_to_jtstype(dtype):
    # Convert
    if pdc.is_bool_dtype(dtype):
        return 'boolean'
    elif pdc.is_integer_dtype(dtype):
        return 'integer'
    elif pdc.is_numeric_dtype(dtype):
        return 'number'
    elif pdc.is_datetime64_any_dtype(dtype):
        return 'datetime'
    else:
        return 'string'



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import collections
import pandas as pd
import jsontableschema
from jsontableschema import Schema
from . import mappers


# Module API

class Storage(object):
    """Pandas Tabular Storage.

    It's an implementation of `jsontablescema.Storage`.

    Args:
        dataframes (list): list of storage dataframes

    """

    # Public

    def __init__(self, dataframes=None):
        self.__dataframes = dataframes or collections.OrderedDict()
        self.__descriptors = {}

    def __repr__(self):
        return 'Storage'

    def __getitem__(self, key):
        return self.__dataframes[key]

    @property
    def buckets(self):
        return list(self.__dataframes.keys())

    def create(self, bucket, descriptor, force=False):

        # Make lists
        buckets = bucket
        if isinstance(bucket, six.string_types):
            buckets = [bucket]
        descriptors = descriptor
        if isinstance(descriptor, dict):
            descriptors = [descriptor]

        # Check buckets for existence
        for bucket in buckets:
            if bucket in self.buckets:
                if not force:
                    raise RuntimeError('Bucket "%s" already exists' % bucket)
                self.delete(bucket)

        # Define dataframes
        for bucket, descriptor in zip(buckets, descriptors):
            jsontableschema.validate(descriptor)
            self.__descriptors[bucket] = descriptor
            self.__dataframes[bucket] = pd.DataFrame()

    def delete(self, bucket=None, ignore=False):

        # Make lists
        buckets = bucket
        if isinstance(bucket, six.string_types):
            buckets = [bucket]
        elif bucket is None:
            buckets = reversed(self.buckets)

        # Iterate over buckets
        for bucket in buckets:

            # Non existent bucket
            if bucket not in self.buckets:
                if not ignore:
                    raise RuntimeError('Bucket "%s" doesn\'t exist' % bucket)

            # Remove from descriptors
            if bucket in self.__descriptors:
                del self.__descriptors[bucket]

            # Remove from dataframes
            if bucket in self.__dataframes:
                del self.__dataframes[bucket]

    def describe(self, bucket, descriptor=None):

        # Set descriptor
        if descriptor is not None:
            self.__descriptors[bucket] = descriptor

        # Get descriptor
        else:
            descriptor = self.__descriptors.get(bucket)
            if descriptor is None:
                dataframe = self.__dataframes[bucket]
                descriptor = mappers.dataframe_to_descriptor(dataframe)

        return descriptor

    def iter(self, bucket):

        # Check existense
        if bucket not in self.buckets:
            raise RuntimeError('Bucket "%s" doesn\'t exist.' % bucket)

        # Prepare
        descriptor = self.describe(bucket)
        schema = Schema(descriptor)

        # Yield rows
        for pk, row in self.__dataframes[bucket].iterrows():
            rdata = []
            for field in schema.fields:
                if schema.primary_key and schema.primary_key[0] == field.name:
                    rdata.append(field.cast_value(pk))
                else:
                    value = row[field.name]
                    rdata.append(field.cast_value(value))
            yield rdata

    def read(self, bucket):

        # Get rows
        rows = list(self.iter(bucket))

        return rows

    def write(self, bucket, rows):

        # Prepare
        descriptor = self.describe(bucket)
        new_data_frame = mappers.descriptor_and_rows_to_dataframe(descriptor, rows)

        # Just set new DataFrame if current is empty
        if self.__dataframes[bucket].size == 0:
            self.__dataframes[bucket] = new_data_frame

        # Append new data frame to the old one setting new data frame
        # containing data from both old and new data frames
        else:
            self.__dataframes[bucket] = pd.concat([
                self.__dataframes[bucket],
                new_data_frame,
            ])
